require 'net/ldap'
require 'puree'

require 'openaccess/apc/exceptions'
require 'openaccess/apc/version'

# Monkey-patch the Puree person-publications query to support additional
# parameters
module Puree
  module Query
    # Add the creation and modification dates to the selection criteria for
    # publications
    class Person
      # Maps keyword arguments to Pure query parameters
      # The query parameter is either a String (the name of the parameter) or
      # a Proc which accepts the keyword argument name, value and query
      # parameter hash and modifies the query hash as required.
      OPTS = {
        created_end: 'createdDate.toDate',
        created_start: 'createdDate.fromDate',
        modified_end: 'modifiedDate.toDate',
        modified_start: 'modifiedDate.fromDate',
        order_by: proc do |_opt, values, query|
          values = [values] unless values.is_a?(Array)
          values.each_index do |i|
            value = values[i]
            if '+-'.include?(value[0])
              descending = value[0] == '-' ? 'true' : 'false'
              value = value[1..-1]
            else
              descending = 'false'
            end
            param = format('orderBy.property[%d]', i)
            query[param] = value
            param = format('orderBy.property[%d].descending', i)
            query[param] = descending
          end
        end,
        published_end: 'publicationDate.toDate',
        published_start: 'publicationDate.fromDate',
        workflow_states: proc do |_opt, values, query|
          values = [values] unless value.is_a?(Array)
          values.each_index do |i|
            param = format('workflowStates.workflowState[%d].workflowName', i)
            query[param] = 'publication'
            param = format('workflowStates.workflowState[%d]', i)
            query[param] = values[i].to_s
          end
        end
      }.freeze

      # Returns an array of publications for the given person's UUID
      # @param uuid [String] the person UUID
      # @param limit [Integer] the number of results to return
      # @param opts [Hash] the options hash
      # @option opts [String] :uuid the person UUID owning the publications
      # @option opts [Numeric] :limit the maximum number of results to return
      # @option opts [Numeric] :offset the start of the result set
      # @option opts [String] :created_start the creation start date
      # @option opts [String] :created_end the creation end date
      # @option opts [String] :modified_start the modification start date
      # @option opts [String] :modified_end the modification end date
      # @option opts [Array<String>] :order_by the list of fields to sort the
      #   result set. Each field can be prefixed by '+' (ascending sort) or '-'
      #   (descending sort); the default is ascending.
      # @option opts [String] :published_start the publication start date
      # @option opts [String] :published_end the publication end date
      # @option opts [Array<String>] :workflow_states the valid workflow
      #   states (:approved|:entryInProgress|:forApproval|:forRevalidation)
      # @return [Array<Puree::Model::Publication] the array of publications
      def publications_extended(uuid:, limit:, **opts)
        uuids = publication_uuids_extended(uuid: uuid, limit: limit, **opts)
        publications = []
        uuids.each do |u|
          publication_extractor = Puree::Extractor::Publication.new(@config)
          publication = publication_extractor.find uuid: u
          publications << publication if publication
        end
        publications
      end

      # Returns an array of publication UUIDs for the given person's UUID
      # @param (see #publications_extended)
      # @return [Array<String>] the array of publication UUIDs
      def publication_uuids_extended(uuid:, limit:, **opts)
        params = {
          'associatedPersonUuids.uuid' => uuid,
          'rendering' => :system,
          'window.size' => (limit || 5).to_s,
          'window.offset' => (opts[:offset] || 0).to_s
        }
        self.class.query_params(params, **opts)
        headers
        response = @req.get(build_publication_url, params: params)
        extract_publication_uuids(make_doc(response.body))
      end

      # Adds whitelisted keyword arguments to the supplied query parameter hash
      # @param params [Hash] the query parameters hash
      # @return [void]
      def self.query_params(params, **opts)
        OPTS.each do |opt, param|
          next unless opts[opt]
          if param.is_a?(Proc)
            param.call(opt, opts[opt], params)
          else
            params[param] = opts[opt]
          end
        end
      end
    end
  end
end

module OpenAccess
  # Methods for managing open access article publishing charges (APCs)
  module APC
    # Exceptions
    include Exceptions

    # Module configuration
    @config = {
      ldap: {
        # The LDAP attribute supplying the Pure employee ID
        attribute_pure_id: nil,
        # The LDAP attribute supplying the network username
        attribute_username: nil,
        auth: {
          # The LDAP authentication method
          method: :simple,
          # The LDAP service user account's password
          password: nil,
          # The LDAP service user account's username
          username: nil
        },
        # The base tree of the LDAP directory
        base: nil,
        # The LDAP server hostname
        host: nil,
        # The LDAP server's port
        port: 389
      },
      pure: {
        # The Pure web service user's password
        password: nil,
        # The base URL of the Pure web service
        url: nil,
        # The Pure web service user's username
        username: nil
      }
    }

    @person_extractor = nil
    @person_query = nil

    # Sets the module configuration
    # @return [void]
    # @yield [config] Passes the configuration hash to the block
    # @yieldparam config [Hash] the configuration hash
    def self.configure
      # Get user-supplied configuration from the block
      yield(@config)

      # LDAP client for person resolution
      @ldap = configure_ldap

      # Pure clients
      @person_extractor = Puree::Extractor::Person.new(@config[:pure])
      @person_query = Puree::Query::Person.new(@config[:pure])
    end

    # Returns the Pure person UUID given a network username
    # @param username [String] the network username
    # @param employee_id [Integer, String] the Pure employee ID
    # @return [String] the Pure UUID of the corresponding person
    def self.person(username = nil, employee_id: nil)
      employee_id ||= person_employee_id(username)
      person = @person_extractor.find_by_id(employee_id: employee_id)
      return person.uuid unless person.nil?
      raise UserNotFound, 'No user for employee ID: ' \
                          "#{employee_id} (username: #{username})"
    end

    # Returns an array of publication summary details for the N most recent
    # funding-eligible publications
    # @param username [String] the network username
    # @param opts [Hash] the options hash
    # @option opts [Date, DateTime, String, Time] :created_end include
    #   publications deposited on or up to this date
    # @option opts [Date, DateTime, String, Time] :created_start include
    #   publications deposited on or after this date
    # @option opts [Integer, String] the Pure employee ID
    # @return [Array<Hash>] the publications
    def self.publications_for_person(username = nil, **opts)
      person_uuid = person(username, employee_id: opts[:employee_id])
      publications_criteria(opts)
      pubs = @person_query.publications_extended(uuid: person_uuid, **opts)
      publications_summary(pubs)
    end

    # Private helpers

    # Returns a configured Net::LDAP instance
    # @return [Net::LDAP] the LDAP instance
    def self.configure_ldap
      conf = @config[:ldap]
      return nil unless conf
      Net::LDAP.new(host: conf[:host], port: conf[:port], auth: conf[:auth])
    end
    private_class_method :configure_ldap

    # Returns the Pure employee ID given a network username
    # @param username [String] the network username
    # @return [String] the Pure employee ID
    def self.person_employee_id(username = nil)
      # Get the LDAP record for the username
      c = @config[:ldap]
      filter = Net::LDAP::Filter.eq(c[:attribute_username], username)
      # Return the value of the LDAP attribute containing the Pure employee ID
      @ldap.search(base: c[:base], filter: filter) do |entry|
        pure_id = entry[c[:attribute_pure_id]]
        return pure_id[0] unless pure_id.nil? || pure_id.empty?
      end
      raise UserNotFound, "No user for username: #{username}"
    end
    private_class_method :person_employee_id

    # Returns a hash of publication summary details ready for JSON formatting
    # @param pub [Puree::Model::Publication] the publication
    # @return [Hash] the publication summary
    # rubocop:disable Metrics/MethodLength
    def self.publication_summary(pub)
      {
        created: pub.created,
        description: pub.description,
        dois: pub.dois,
        subtitle: pub.subtitle,
        title: pub.title,
        translated_subtitle: pub.translated_subtitle,
        translated_title: pub.translated_title,
        uuid: pub.uuid,
        workflow_state: pub.workflow_state
      }
    end
    # rubocop:enable Metrics/MethodLength
    private_class_method :publication_summary

    # Sets the Puree person query criteria in an options hash
    # @param opts [Hash] the options hash
    # @return [Hash] the options hash
    def self.publications_criteria(opts)
      opts[:limit] ||= 5
      opts[:order_by] = ['-created']
      end_date = to_date(opts[:created_date_end])
      start_date = to_date(opts[:created_date_start])
      opts[:created_end] = end_date.to_s if end_date # YYYY-MM-DD
      opts[:created_start] = start_date.to_s if start_date
      opts
    end
    private_class_method :publications_criteria

    # Returns a list of publication summaries ready for JSON formatting
    # @param pubs [Array<Puree::Model::Publication>] the publications
    # @return [Array<Hash>] the publication summaries
    def self.publications_summary(pubs)
      pubs.map { |p| publication_summary(p) }
    end
    private_class_method :publications_summary

    # Converts a date representation to a Date instance.
    # Date instances are returned as-is. DateTime and Time instances are
    # converted to Date instances. All other types are converted to a string
    # and parsed.
    # @param date [Object] the date representation
    # @param default [Date, nil] the default value if no date is supplied or
    #   string parsing fails
    # @return [Date] the date instance
    def self.to_date(date = nil, default = nil)
      return default if date.nil? || date.empty?
      return date if date.is_a?(Date)
      return date.to_date if date.respond_to?(:to_date)
      Date.parse(date.to_s)
    rescue ArgumentError
      default
    end
    private_class_method :to_date
  end
end